function [bX bH bHX bHH logZ] = HCRFInferExact(F, G, E, W, V, X, use_ve, enums)
%Inferences on HCRF by sampling
%nX: the number of label states
%nH: the number of hidden states
%F: the observation for association terms. dA*nNode
%G: the observation for interaction terms. dI*nEdge
%E: the edges between nodes. its size should be 2*nEdge. no symmetric edges
%   shoudl be contained.
%W: the coefficient vectors for association terms. its size should be d1*nH
%V: the coefficient vectors for interaction terms. its size should be d2*nH*nH
%X: the belief vector of nodes. its size should be nX*nNode. should be [] if testing

%bX: the beliefs of X node. its size should be nX*nNode
%bH: the beliefs of H node. its size should be nH*nNode
%bHH: the belief of H edge. its size should be (nH^2)*nEdgeEx
%bXH: the beliefs of X-H edges. its size should be (nH*nX)*nNode
%logZ: the value of partition function. scalar.

[dA nH nX] = size(W);
dI = size(V, 1);
nNode = size(F, 1);
nEdge = size(G, 1);
isTrain = ~isempty(X);

FW = reshape(F*reshape(W, [dA, nH*nX]), [nNode, nH, nX]);

if nargin < 8
    enums = [];
    if nargin < 7
        use_ve = nEdge == 0 | all(V == 0);
    end
end

if use_ve % variable elimination (BP)
    assert(nEdge == 0 | all(V(:) == 0), 've can only be used without edges');
    if isTrain
        FW = FW(:,:,X);
        mp = 0.5*(max(FW,[],2) + min(FW,[],2));
        FW = bsxfun(@plus, FW, -mp);
        eFW = exp(FW);
        logZ = log(prod(sum(eFW, 2), 1)) + sum(mp);
        
        bX = zeros(1, nX); 
        bX(X) = 1;
        bHX = zeros(nNode, nH, nX); 
        bHX(:,:,X) = bsxfun(@times, eFW, 1./sum(eFW,2));
    else
        FW = reshape(FW, nNode, nH*nX);
        mp = 0.5*(max(FW,[],2) + min(FW,[],2));
        FW = reshape(bsxfun(@plus, FW, -mp), [nNode, nH, nX]);
        eFW = exp(FW);
        msgX = sum(eFW, 2);
        bX = prod(msgX, 1);
        Z = sum(bX);
        logZ = log(Z) + sum(mp);
        
        msgH = bsxfun(@times, bX, 1./msgX);
        bHX = bsxfun(@times, eFW, msgH)./Z;
        bX = bX(:)'./Z;
    end
    bH = sum(bHX, 3);
    bHH = [];
else % enumeration
    assert(nX*nH^nNode <= 1e7,'too many samples');

    if ~isempty(enums)
        if isTrain
            samples = enums{nNode}(1:(end - 1), 1:(nH^nNode));
            samples = [samples; repmat(X,1,size(samples, 2))];
        else
            samples = enums{nNode};
        end
    else
        if isTrain
            samples = GetCombinations(repmat(nH,1,nNode));
            samples = [samples; repmat(X,1,size(samples, 2))];
        else
            samples = GetCombinations([repmat(nH,1,nNode), nX]);
        end
    end
    samples = int32(samples) - 1;

    if nEdge > 0
        GV = reshape(G*reshape(V, [dI, nH*nH]), [nEdge, nH, nH]);
    else
        GV = [];
    end

    P = hcrf_phi(FW, GV, int32(E) - 1, samples);

    mp = max(P);
    P = exp(P - mp);
    Z = sum(P);
    P = P./Z;
    logZ = log(Z) + mp;

    %[ bX bH bHX bHH ] = HCRFGetB(nH, nX, E, samples, P);
    [bX bHX bHH] = hcrf_getb( nH, nX, int32(E) - 1, samples, P );
    bH = sum(bHX, 3);
end
