classdef GMM < handle
% gaussian mixture  model

    properties
        weights
        means
        covars
    end
    
    methods
        function [model] = GMM(weights, means, covars)
            % [model] = GMM(weights, means, covars)
            % weights: the weights of components
            % means: the mean of Gaussians, one column for one component
            % covars: the covariances of Gaussians, one slice for one component
            
            if nargin > 0
                assert(length(weights) == size(means,2), ...
                    'number of components should match');

                model.weights = weights; % normalize to one
                model.means = means;

                if isscalar(covars) % uniform iso covariance
                    covars = repmat(covars*eye(model.Dim()), [1,1,length(weights)]);
                elseif ndims(covars) == 1
                    assert(length(weights) == length(covars),...
                        'number of components should match');

                    tmp = zeros(model.Dim(), model.Dim(), model.N());
                    for ind = 1:model.N(); 
                        tmp(:, :, ind) = eye(model.Dim())*covars(ind); 
                    end
                    covars = tmp;
                elseif ndims(covars) == 2
                    assert(size(covars,1) == size(covars,2) && size(covars,1) == self.Dim(), ...
                        'dim should match');

                    covars = repmat(covars, [1, 1, 3]);
                end

                model.covars = covars;
            end
        end
        
        function [n] = N(self)
            % [n] = N(self)
            % number of components
            
            n = length(self.weights);
        end
        
        function [dim] = Dim(self)
            % [dim] = Dim(self)
            % dimension of the data
            
            dim = size(self.means, 1);
        end
        
        function [data, labels] = GenData(self, n)
            % [data, labels] = GenData(self, n)
            % n: number of points
            % data: generated points. one row per point
            % labels: which components generated the points
            
            labels = randm(self.weights/sum(self.weights), n);
            
            data = zeros(n, self.Dim());
            for l = 1:self.N()
                idx = labels == l;
                data(idx, :) = mvnrnd(self.means(:, l), self.covars(:,:,l), sum(idx));
            end
        end
        
        function [] = Train(self, X, K, options)
            % [] = Train(X, K)
            % training this gmm from data
            
            if nargin < 4
                options = [];
            end

            [ntry, maxIter] = GetOptions(options, 'ntry', 5, 'maxIter', 100);
            
            model = gmdistribution.fit(X, K, 'Start', 'randSample', ...
                'Replicates', ntry,'Regularize', 1e-5,'Options',struct('MaxIter', maxIter));
            
            self.weights = model.PComponents;
            self.means = model.mu';
            self.covars = model.Sigma;
        end
        
        function [L] = Likelihood(self, X)
            % [L] = Likelihood(self, X)
            % compute the log-likelihood of each point under this model
            
            K = length(self.weights);
            lnpdf = zeros(size(X, 1), K);
            for k = 1:K
                lnpdf(:, k) = mvnlnpdf(X, self.means(:,k)', self.covars(:,:,k));
            end

            mp = max(lnpdf, [], 2);
            L = log(exp(bsxfun(@plus, lnpdf, -mp))*self.weights(:) + eps(0.0)) + mp;
        end
        
        function [new] = Clone(self, with_content)
            %[new] = Clone(self, with_content)
            % clone the current model
            
            if nargin < 2
                with_content = true;
            end
            
            new = GMM();
            if with_content
                ps = properties(self);
                for i = 1:length(ps)
                    new.(ps{i}) = self.(ps{i});
                end
            end
        end
        
        function [] = Plot(self)
            % [] = Plot(self)
            % plot points from this model
            
            assert(self.Dim() <= 2, 'can only plot in 2-D space');
            
            [data, labels] = self.GenData(1000);
            if self.Dim() == 2
                scatter(data(:,1), data(:,2), [], labels, 'marker', '.');                
                axis equal;
            else
                hist(data, 20); hold on;
                scatter(data, -2*ones(1, length(data)), [], labels, 'marker', 'o');hold off;
            end
        end
        
        function [] = Contour(self)
            % [] = Draw(self)
            % draw the contour plot of the current model's density
            
            assert(self.Dim() == 2, 'can only plot in 2-D space');
            
            gm = gmdistribution(self.means', self.covars, self.weights./sum(self.weights));
            r = MinMax(self.means, 2);
            d = zeros(size(r));
            d(:,1) = r(:,1) - (r(:,2) - r(:,1));
            d(:,2) = r(:,2) + (r(:,2) - r(:,1));
            ezcontour(@(x,y) pdf(gm, [x y]), d(1,:), d(2,:));
        end
    end
end
