classdef PCA < handle
    
    properties
        U % basis
        L % eigenvalues
        M % mean
        R % effective rank
        E % energy preserved
    end
    
    methods
        function [model] = PCA(X, E)
            if nargin > 0
                if nargin < 2
					E = 0.99;
				end
                model.Train(X, E);
            end
        end
        
        function Train(self, X, E)
            if nargin < 3
				E = 0.99;
			end
            
            [n, dim] = size(X);
            self.R = min(n, dim);
            self.M = mean(X, 1);
            X = bsxfun(@plus, X, -self.M);
            
            if n >= dim
                C=X'*X./n;
                [self.U S]=eig(C);
                [self.L I]=sort(diag(S), 'descend');
                self.U=self.U(:, I);
            else
                C=X*X'./n;
                [V S]=eig(C);
                [self.L I]=sort(diag(S), 'descend');
                self.U=X'*V(:, I)*diag(1./self.L./sqrt(n));
            end
            
            self.U = Normalize(self.U, 'n1');
            self.E = E;
            self.R = EffRank(self.L, self.E);
        end
        
        function [P] = Project(self, X, end_dim, start_dim)
			if ~exist('end_dim', 'var') || isempty(end_dim)
				end_dim = self.R;
			end
			if ~exist('start_dim', 'var') || isempty(start_dim)
				start_dim = 1;
			end

            P=bsxfun(@plus, X, -self.M)*self.U(:, start_dim:end_dim);
        end

        function [X] = Reconstruct(self, P)
            r = size(P, 2);
            X = bsxfun(@plus, P*self.U(:, 1:r)', self.M);
        end
    
        function [h] = Visualize(self, X, c)
            if nargin < 3
                c = 'b';
            end
            
            P = self.Project(X, 2);
            h = scatter(P(:,1), P(:,2), [], c, 'marker', '.');
        end
        
        function [h] = Visualize3(self, X, c)
            if nargin < 3
                c = 'b';
            end
            
            P = self.Project(X, 3);
            h = scatter3(P(:,1), P(:,2), P(:, 3), [], c, 'marker', '.');
        end
    
        function [] = Plot(self, pcs)
            if nargin < 2
                pcs = 1:2;
            end
            
            subplot(2,2,1);
            loglog(1:self.R, self.L(1:self.R), 'b', self.R:length(self.L), self.L(self.R:length(self.L)), 'r');
            title('Eigenvalues');
            subplot(2,2,2);plot(self.M);title('Data mean');
            subplot(2,2,3);plot(self.U(:,pcs(1)));title('First Basis');
            subplot(2,2,4);plot(self.U(:,pcs(2)));title('Second Basis');
            drawnow;
        end
    end
end
